// bslide -- main window
// (C)Copyright 2004, Yashiro Takeshi

#include <btron/dp.h>
#include <btron/hmi.h>
#include <btron/libapp.h>
#include <keycode.h>
#include <tcode.h>
#include <bstdio.h>
#include "BSlideWindow.h"
#include "LinkCollector.h"
#include "BSlideJumpPanel.h"

extern DEV_SPEC SCREEN;

BSlideWindow::BSlideWindow(LINK* link = 0)
: bsfile(link), link(link), wid(-1), title((TC[]){TCS_BSLIDE}), currentSlide(-1)
{
  bgpat.spat.kind = 0;
  bgpat.spat.hsize = bgpat.spat.vsize = 1;
  bgpat.spat.fgcol = 0x90000000;
  bgpat.spat.bgcol = 0x90ffffff;
  bgpat.spat.mask = FILL50;
  // needs to getscreen() beforehand in main.cc
  mgid = create_memgid(&screen_bmp, SCREEN.hpixels, SCREEN.vpixels);
  if (mgid < 0){
    throw BException((TC[]){TCS_ERROR_MEMGID});
  }
  setrect(wrect, 0, 0, SCREEN.hpixels, SCREEN.vpixels);
  jumpSlide(0);
}

// ɽڤؤ(ºݤˤϥɥΥץ󡦥)
void BSlideWindow::display(bool state = true)
{
  if (state){
    if (wid < 0){
      wid = wopn_wnd(WA_SFULL, 0, &wrect, NULL, 2, title, &bgpat, NULL);
      wreq_dsp(wid);
    }
  }
  else {
    if (wid >= 0){
      wcls_wnd(wid, CLR);
      wid = -1;
    }
  }
}

// ɽ
void BSlideWindow::redisp(WEVENT* ev)
{
  if (wid < 0){
    throw BException((TC[]){TCS_ERROR_BADWID});
  }
  do {
    if (wsta_dsp(wid, NULL, NULL) == 0)
      break;
    wera_wnd(wid, NULL);    
    GID gid = wget_gid(wid);
    grst_bmp(gid, &(screen_bmp.bounds), &screen_bmp, &(screen_bmp.bounds),
             NULL, G_STORE);
  } while (wend_dsp(wid) > 0);
  return;
}

// 
void BSlideWindow::onMouseButtonDown(WEVENT* ev)
{
  switch (ev->s.cmd) {
  case W_PICT:
    switch (wchk_dck(ev->s.time)) {
    case W_DCLICK:
      terminate();
    default:
      return;
    }
  case W_WORK:
    nextSlide();
    return;
  default:
    return;
  }
  return;
}

// Ͻ
void BSlideWindow::onKeyDown(WEVENT* ev)
{
  if (ev->s.stat & ES_CMD){
    W i = mfnd_key(menu.getMenuID(), ev->s.wid);
    if (i > 0){
      selectMenu(ev, i);
    }
    return;
  }
  // ʲ̾省ν
  switch (ev->e.data.key.code){
  case KC_CC_U: case KC_CC_L: case KC_PG_U:
  case KC_SC_U: case KC_SC_L:
  case TK_P:    case TK_p:
    prevSlide();
    break;
  case KC_CC_D: case KC_CC_R: case KC_PG_D:
  case KC_SC_D: case KC_SC_R: case KC_SPACE:
  case TK_N:    case TK_n:
    nextSlide();
    break;
  }
  return;
}

// ˥塼ܵư
void BSlideWindow::selectMenu(WEVENT *ev, W mno)
{
  BSlideMenuItem mitem = menu.getInternalID(mno);
  
  switch (mitem){
  case MN_FINISH: // λ
    terminate();
    break;
  case MN_REDISP: // ɽ
    wreq_dsp(wid);
    break;
  case MN_GONEXT: // ؿʤ
    nextSlide();
    break;
  case MN_GOBACK: // 
    prevSlide();
    break;
  case MN_JUMP:   // 
    {
      BSlideJumpPanel pnl;
      pnl.setFileList(link);
      pnl.invalidateFileList();
      pnl.setSelection((currentSlide >= link.size()) ? 0 : currentSlide);
      if (pnl.execute() == 1){
        jumpSlide(pnl.getSelection());
      }
    }
    break;
  case MN_WINDOW: // ɥ
    wexe_dmn(mno);
    break;
  case MN_GADGET: // ʪ
    oexe_apg(0, mno);
    break;
  }
}

// ɥν
void BSlideWindow::onIdle(WEVENT *ev)
{
  if (ev->s.wid != wid ||
      ev->s.cmd != W_WORK ||
      ev->s.stat & ES_CMD){
    return;
  }
  static PNT prev = {-1, -1};
  static UW ptime = ev->s.time;
  bool showing = true;
  if (prev.x != ev->s.pos.x || prev.y != ev->s.pos.y){
    ptime = ev->s.time;
    prev.x = ev->s.pos.x;
    prev.y = ev->s.pos.y;
    showing = true;
  }
  if (!showing || ev->s.time - ptime > 3000){
    showing = false;
  }
  setpointer(PS_SELECT, NULL);
  gdsp_ptr(showing ? 1 : 0);
}

// ٥Ƚμ¹
void BSlideWindow::run()
{
  display(true);
  
  WEVENT wev0; 
  // ٥ȥ롼: 㳰ȤΤǼ
  for (cont = true; cont; ) {
    wget_evt(&wev0, WAIT);
    
    switch (wev0.s.type){
    case EV_NULL:
      onIdle(&wev0);
      break;
    case EV_REQUEST:
      switch (wev0.g.cmd) {
      case W_REDISP:
        redisp(&wev0);
        break;
      case W_PASTE:
        wrsp_evt(&wev0, 1); // NACK
        break;
      case W_DELETE:
      case W_FINISH:
        wrsp_evt(&wev0, 0); // ACK
        terminate();
      }
      break;
    case EV_RSWITCH:
      redisp(&wev0);
    case EV_SWITCH:
    case EV_BUTDWN:
      onMouseButtonDown(&wev0);
      break;
    case EV_INACT:
      break;
    case EV_DEVICE:
      oprc_dev(&(wev0.e), NULL, 0);
      break;
    case EV_MSG:
      clr_msg(MM_ALL, MM_ALL);
      break;
    case EV_MENU:
      {
        W mi = menu.popupMenu(&wev0);
        if (mi > 0){
          selectMenu(&wev0, mi);
        }
      }
      break;
    case EV_KEYDWN:
      onKeyDown(&wev0);
      break;
    }
  }
  
  display(false);
}

void BSlideWindow::jumpSlide(int index)
{
  if (index < 0 || index > link.size()){
    throw BException((TC[]){TCS_ERROR_OUTRANGE});
  }
  if (currentSlide == index){
    return;
  }
  currentSlide = index;

  RECT rect = {{ 0, 0, SCREEN.hpixels, SCREEN.vpixels }};
  
  if (index == link.size()){
    // λľ
    PAT pat = {{ 0, 1, 1, 0x10000000, 0x10ffffff, FILL100 }};
    SIZE fsize = {16, 16}; 
    FSSPEC fnt = {{TCS_FONTNAME}, 0x18c0, FT_PROP|FT_GRAYSCALE, fsize};
    gfil_rec(mgid, rect, &pat, 0, G_STORE);
    gset_chc(mgid, 0x10ffffff, 0x10000000);
    gset_fon(mgid, &fnt);
    gdra_stp(mgid, 16, 16+16, (TC[]){TCS_FINISH_BSLIDE},
      sizeof((TC[]){TCS_FINISH_BSLIDE})/sizeof(TC), G_STORE);
    wreq_dsp(wid);
    return;
  }
  
  link.drawObject(mgid, currentSlide, rect);
  wreq_dsp(wid);
  return;
}

