// bslide -- iteration of linked segments+records
// (C)Copyright 2004, Yashiro Takeshi

#include <new>
#include "LinkCollector.h"
#include "Exceptions.h"
#include "tcsliteral.h"
#include <btron/file.h>
#include <bstdio.h>

extern DEV_SPEC SCREEN;

// 󥯥쥳ɤɤ߹
void LinkCollector::readLinkRecords(W fd)
{
  W err, recnum, size, rectype;
  UH subtype;
  bool first = true;
  LINK link;
  
  do {
    err = fnd_lnk(fd, (first ? F_TOPEND : F_NFWD), NULL, 0, &recnum);
    first = false;
    if (err < 0)
      break;
    rectype = rea_rec(fd, 0, (B*)&link, sizeof(LINK), &size, &subtype);
    if (rectype < 0 || size != sizeof(LINK)){
      // ɤ߹ߤ˼ / 󥯥쥳ɤ
      cls_fil(fd);
      throw BException((TC[]){TCS_ERROR_FOPEN});
    }
    linkobj.push_back(link);
  } while (err >= 0);
  return;
}  

// 󥯥Ȥɤ߹
void LinkCollector::readLinkSegments(W fd)
{
  W size, recnum, err, rectype;
  UH subtype;
  B* buf = 0;
  
  try {
    err = fnd_rec(fd, F_TOPEND, (1 << RT_TADDATA), 0, &recnum);
    if (err < 0){
      throw BException((TC[]){TCS_ERROR_RECORD});
    }
    rectype = rea_rec(fd, 0, NULL, 0, &size, &subtype);
    buf = new(std::nothrow) B[size];
    if (buf == 0){
      // γݤ˼
      throw BException((TC[]){TCS_ERROR_BADALLOC});
    }
    err = rea_rec(fd, 0, buf, size, NULL, NULL);
    
    // parse TAD here
    for (int pos = 0; pos < size; ){
      if (buf[pos+1] == '\xff' && (buf[pos] & 0x80) != 0){
        if (((TADSEG*)buf)->len == 0xffffU){
          pos += ((LTADSEG*)buf)->llen + sizeof(LTADSEG);
        }
        else {
          int tsize = ((TADSEG*)&buf[pos])->len;
          if (buf[pos] == '\xe6'){ // TS_VOBJ
            B* tmp = new(std::nothrow) B[tsize];
            if (tmp == 0){
              // γݤ˼
              throw BException((TC[]){TCS_ERROR_BADALLOC});
            }
            memcpy(tmp, &buf[pos+sizeof(TADSEG)], tsize);
            linkseg.push_back((VOBJSEG*)tmp);
          }
          pos += tsize + sizeof(TADSEG);
        }
      }
      else {
        opts.putChar(*((TC*)&buf[pos]));
        pos += sizeof(TC);
      }
    }
    // %size:... ץμ
    std::string sizestr = opts.getOption("size");
    if (!sizestr.empty()){
      orgsize = opts.parseSizeOption(sizestr);
      //printf("%d %d\n", orgsize.h, orgsize.v);
    }
  }
  catch (...){
    for (int i = 0; i < linkseg.size(); i++){
      delete [] ((B*)linkseg[i]);
    }
    delete [] buf;
    throw;
  }
  delete [] buf;
}

LinkCollector::LinkCollector(LINK* parent = 0)
: parent(parent)
{
  W fd = -1;
  B* buf = 0;
  orgsize.h = orgsize.v = -1;
  
  try {
    fd = opn_fil(parent, F_READ, NULL);
    if (fd <= 0){
      // եΥץ˼
      throw BException((TC[]){TCS_ERROR_FOPEN});
    }
    readLinkRecords(fd);  
    readLinkSegments(fd);
    //printf("records:  %d\n", linkobj.size());
    //printf("segments: %d\n", linkseg.size());
    int diff = linkobj.size() - linkseg.size();
    for (int i = 0; i < diff; i++){
      VOBJSEG* vs = (VOBJSEG*)(new(std::nothrow) B[sizeof(VOBJSEG)]);
      if (vs == 0){
        for (int j = 0; j < i; j++){
          delete [] (B*)(linkseg[linkseg.size()+j]);
        }
        throw BException((TC[]){TCS_ERROR_BADALLOC});
      }
      
      setrect(vs->view, 0, 0, SCREEN.hpixels, SCREEN.vpixels);
      vs->chsz = 0x0010;
      vs->frcol = vs->chcol = 0x10000000;
      vs->tbcol = 0x10dfdfdf;
      vs->bgcol = 0x10ffffff;
      vs->dlen = 0;
      linkseg.push_back(vs);
    }
  }
  catch (...){
    if (fd > 0){
      cls_fil(fd);
    }
    throw;
  }
  cls_fil(fd);
}

// ȤȤ̵褹
void LinkCollector::drawObject(GID gid, int n, const RECT& rect)
{
  if (n < 0 || n >= size()){
    throw BException((TC[]){TCS_ERROR_OUTRANGE});
  }
  
  setrect(linkseg[n]->view, 0, 0, SCREEN.hpixels, SCREEN.vpixels);
  linkseg[n]->height = SCREEN.vpixels;
  VLINK *p = reinterpret_cast<VLINK*>(&linkobj[n]);
  p->attr &= 0x3bdfU; // 0011 1011 1101 1111
  p->attr |= 0x01cfU; // 0000 0001 1100 1111
  
  if (orgsize.h < 0 || orgsize.v < 0 ||
      (SCREEN.hpixels == orgsize.h && SCREEN.vpixels == orgsize.v)){
    odra_vob(p, linkseg[n], -gid, V_DISPALL);
  }
  else {
    setrect(linkseg[n]->view, 0, 0, orgsize.h, orgsize.v);
    PNT scal =
      { 256 * SCREEN.hpixels / orgsize.h, 
        256 * SCREEN.vpixels / orgsize.v };
    odra_vor(p, linkseg[n], -gid, V_DISPALL, scal);
  }
}

void LinkCollector::getName(TC* name, int n) const
{
  if (n < 0 || n >= size()){
    throw BException((TC[]){TCS_ERROR_OUTRANGE});
  }
  fil_sts(const_cast<LINK*>(&linkobj[n]), name, NULL, NULL);
}
